/*
 * Copyright (C) 2025 Dominik Drexler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef TYR_FORMALISM_PLANNING_UNARY_OPERATOR_DATA_HPP_
#define TYR_FORMALISM_PLANNING_UNARY_OPERATOR_DATA_HPP_

#include "tyr/common/types.hpp"
#include "tyr/common/types_utils.hpp"
#include "tyr/formalism/planning/declarations.hpp"
#include "tyr/formalism/planning/function_expression_data.hpp"
#include "tyr/formalism/planning/ground_function_expression_data.hpp"
#include "tyr/formalism/planning/unary_operator_index.hpp"

namespace tyr
{
template<formalism::OpKind Op, typename T>
struct Data<formalism::planning::UnaryOperator<Op, T>>
{
    using OpType = Op;

    Index<formalism::planning::UnaryOperator<Op, T>> index;
    T arg;

    Data() = default;
    Data(Index<formalism::planning::UnaryOperator<Op, T>> index, T arg) : index(index), arg(arg) {}
    Data(const Data& other) = default;
    Data& operator=(const Data& other) = default;
    Data(Data&& other) = default;
    Data& operator=(Data&& other) = default;

    void clear() noexcept
    {
        tyr::clear(index);
        tyr::clear(arg);
    }

    auto cista_members() const noexcept { return std::tie(index, arg); }
    auto identifying_members() const noexcept { return std::tie(Op::kind, arg); }
};

static_assert(!uses_trivial_storage_v<formalism::planning::UnaryOperator<formalism::OpAdd, Data<formalism::planning::FunctionExpression>>>);

}

#endif
