/*
 * Copyright (C) 2025 Dominik Drexler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef TYR_COMMON_COMMON_HPP_
#define TYR_COMMON_COMMON_HPP_

#include "tyr/common/bit.hpp"
#include "tyr/common/bit_packed_array_pool.hpp"
#include "tyr/common/bit_packed_array_set.hpp"
#include "tyr/common/block_array_pool.hpp"
#include "tyr/common/block_array_set.hpp"
#include "tyr/common/chrono.hpp"
#include "tyr/common/closed_interval.hpp"
#include "tyr/common/config.hpp"
#include "tyr/common/declarations.hpp"
#include "tyr/common/dynamic_bitset.hpp"
#include "tyr/common/equal_to.hpp"
#include "tyr/common/formatter.hpp"
#include "tyr/common/hash.hpp"
#include "tyr/common/index_mixins.hpp"
#include "tyr/common/indexed_hash_set.hpp"
#include "tyr/common/itertools.hpp"
#include "tyr/common/memory.hpp"
#include "tyr/common/observer_ptr.hpp"
#include "tyr/common/raw_array_pool.hpp"
#include "tyr/common/raw_array_set.hpp"
#include "tyr/common/segmented_vector.hpp"
#include "tyr/common/types.hpp"
#include "tyr/common/uint_mixins.hpp"
#include "tyr/common/unordered_set.hpp"
#include "tyr/common/variant.hpp"
#include "tyr/common/vector.hpp"

#endif